package com.boot.bus.sqlsync.listener;

import com.alibaba.fastjson.JSON;
import com.alibaba.fastjson.JSONObject;
import com.boot.bus.sqlsync.async.executor.Async;
import com.boot.bus.sqlsync.async.parallel.ParWorker;
import com.boot.bus.sqlsync.async.wrapper.WorkerWrapper;
import com.boot.bus.sqlsync.service.infc.IContingencyPlanInstance;
import com.yeejoin.amos.fas.common.enums.DataSyncOperationEnum;
import com.yeejoin.amos.fas.common.enums.DataSyncTypeEnum;
import com.yeejoin.amos.fas.dao.entity.ContingencyPlanInstance;
import com.yeejoin.amos.fas.datasync.DataSyncMessage;
import lombok.extern.slf4j.Slf4j;
import org.eclipse.paho.client.mqttv3.IMqttMessageListener;
import org.eclipse.paho.client.mqttv3.MqttMessage;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Component;

import java.io.Serializable;
import java.util.*;
import java.util.concurrent.LinkedBlockingDeque;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.stream.Collectors;

/**
 * <h1></h1>
 *
 * @Author SingleTian
 * @Date 2021-04-02 09:22
 */
@Slf4j
@Component
public class SyncMqttMessageListener implements IMqttMessageListener {

    final Deque<String> redisKeysQueue1 = new LinkedBlockingDeque<>();
    final Deque<String> redisKeysQueue2 = new LinkedBlockingDeque<>();
    final AtomicBoolean flag = new AtomicBoolean(true);

    @Autowired
    private IContingencyPlanInstance contingencyPlanInstance;

    @Autowired
    private RedisTemplate redisTemplate;

    private Queue<String> getUsedQueue() {
        return flag.get() ? redisKeysQueue1 : redisKeysQueue2;
    }

    private Queue<String> getUnusedQueue() {
        return flag.get() ? redisKeysQueue2 : redisKeysQueue1;
    }

    @Override
    public void messageArrived(String topic, MqttMessage mqttMessage) throws Exception {
        DataSyncMessage message = null;
        try {
            message = DataSyncMessage.bytes2Message(mqttMessage.getPayload());
            System.out.println(JSON.toJSONString(message));
            log.info("topic:{},message:{}", topic, message);
            ParWorker w = new ParWorker();
            WorkerWrapper<DataSyncMessage, String> workerWrapper = new WorkerWrapper.Builder<DataSyncMessage, String>()
                    .worker(w)
                    .callback(w)
                    .param(message)
                    .build();
            Async.beginWork(200, workerWrapper);
//            syncData(message);
        } catch (Exception e) {
            log.error("同步mqtt推送数据出错:topic:{},message:{},错误信息：{}", topic, message, e.toString());
        } finally {
            assert message != null : "";
            // 待删除的redis的key值放入队列
            getUsedQueue().offer(message.redisKey());
        }
    }

    /**
     * 尝试删除redis中mqtt已经发送成功的同步数据
     * 每10秒轮循一次
     */
    @Scheduled(fixedRate = 10 * 1000)
    synchronized private void cleanRedisKeysFinished() {
        Queue<String> usedQueue = getUsedQueue();
        Queue<String> unusedQueue = getUnusedQueue();
        flag.set(!flag.get());
        while (!usedQueue.isEmpty()) {
            String key = usedQueue.poll();
            if (!redisTemplate.delete(key)) {
                unusedQueue.offer(key);
            }
        }
    }

    /**
     * 处理redis中存在但mqtt未接收到的同步数据
     * 每小时执行一次，整点触发
     */
    @Scheduled(cron = "0 0 * * * ?")
    synchronized private void syncByRedis() {
        Set<String> keys = new HashSet<>();
        for (DataSyncTypeEnum typeEnum : DataSyncTypeEnum.values()) {
            keys.addAll(redisTemplate.keys(typeEnum.toString()));
        }
        if (keys.isEmpty()) {
            return;
        }
        // 为减少redis
        new Timer().schedule(new TimerTask() {
            @Override
            public void run() {
                Queue<String> usedQueue = new LinkedBlockingDeque<>(getUsedQueue());
                keys.forEach(key -> {
                    if (!usedQueue.contains(key)) {
                        String messageStr = String.valueOf(redisTemplate.opsForValue().get(key));
                        DataSyncMessage message = null;
                        try {
                            message = JSON.parseObject(messageStr, DataSyncMessage.class);
                            syncData(message);
                        } catch (Exception e) {
                            log.error("同步redis推送数据出错:key:{},message:{},错误信息：{}", key, message, e.getMessage());
                        } finally {
                            getUsedQueue().offer(message.redisKey());
                        }
                    }
                });
            }
        }, 1000 * 30);
    }

    /**
     * @param message
     */
    private void syncData(DataSyncMessage message) {
        DataSyncTypeEnum type = message.getType();
        DataSyncOperationEnum operation = message.getOperation();
        List<Serializable> data = message.getData();
        assert message.getType() != null && message.getOperation() != null : "同步消息体信息不足";
        if (data == null || data.isEmpty()) {
            return;
        }
        switch (type) {
            case CONTINGENCY_PLAN_INSTANCE: {
                switch (operation) {
                    case DELETE: {
                        contingencyPlanInstance.astDeleteByIds(data.stream().map(x -> ((JSONObject) x).toJavaObject(ContingencyPlanInstance.class).getId()).collect(Collectors.toList()));
                        break;
                    }
                    default: {
                        contingencyPlanInstance.astSaveOrUpdateBatch(data.stream().map(x -> ((JSONObject) x).toJavaObject(ContingencyPlanInstance.class)).collect(Collectors.toList()));
                    }
                }
                break;
            }
            default:
        }
    }
}
